#!/usr/bin/python
#
# Copyright 2018, Gurobi Optimization, LLC and End-to-End Analytics, LLC
#
# Author:      Nicole Taheri (nicole@e2eanalytics.com)
# Last edited: 17 July 2018
#
# This sample formulates and solves the following simple MIP model:
#   minimize    sum_{t=1}^T sum_{m=1}^M r_m * s_{t,m}
#   subject to  sum_{t=1}^T s_{t,m} <= 11              for all m  
#               sum_{m=1}^M s_{t,m} >= l_t             for all m  
#               v_{t,m} = max(0, s_{t,m} - s_{t-1,m})  for all m, t
#               sum_{t=1}^T v_{t,m} <= 1               for all m 
#  s, v binary
#
# There are a few added constraints that can be found in the corresponding
# presentation/webinar slides. 
#
#

from gurobipy import *

try:

	# define parameters
	M               = 4                                         # number of burger makers
	T               = 14                                        # number of time periods
	timeperiods     = range(T)                                  # burger makers
	burgermakers    = ['Tom','Bri','Jack','Barb']               # burger makers
	hourly_rates    = {'Tom': 13,'Bri': 16,'Jack':10,'Barb':10} # rates
	minimum_workers = [2,2,2,3,3,3,2,2,3,3,3,2,2,2]             # min num workers every hour
	maximum_hours   = 11                                        # max num hours each worker can work
	
	# Create a new model
	m = Model("diner_schedule")
	
	# Create variables
	w     = m.addVars(burgermakers, timeperiods, lb=-1, ub=1, name="w")
	v     = m.addVars(burgermakers, timeperiods, name="v")
	not_s = m.addVars(burgermakers, timeperiods, name="not_s")
	s     = m.addVars(burgermakers, timeperiods, vtype=GRB.BINARY, name="s")
	man_s = m.addVars(timeperiods, name="man_s")

	# Set objective
	m.setObjective(quicksum([(s[(b,t)] * hourly_rates[b]) for b in burgermakers for t in timeperiods]), GRB.MINIMIZE)
	
	# Add constraints on maximum daily hours
	m.addConstrs((quicksum([s[(b,t)] for t in timeperiods]) <= maximum_hours for b in burgermakers), name="max_daily_hours")
	
	# Add constraints on minimum hourly works
	m.addConstrs((quicksum([s[(b,t)] for b in burgermakers]) >= minimum_workers[t] for t in timeperiods), name="min_workers")
	
	# Add constraints to count shift starts 
	m.addConstrs((w[(b,t)] == (s[(b,t)] - s[(b,t-1)]) for b in burgermakers for t in range(1,T)), name="shift_starts_hat")
	m.addConstrs((v[(b,t)] == max_(w[(b,t)], 0.0) for b in burgermakers for t in range(1,T)), name="shift_starts")
	m.addConstrs((v[(b,0)] == s[(b,0)] for b in burgermakers), name="shift_starts_init")

	# Add constraints to place maximums on shift starts and ends
	# OPTION 1: linear constraints
	m.addConstrs((quicksum([v[(b,t)] for t in timeperiods]) <= 1 for b in burgermakers ), name="shift_start_max")
	# OPTION 2: linear constraints, but created in a loop instead of all at once
	#for b in burgermakers: 
	#	m.addConstr((quicksum([v[(b,t)] for t in timeperiods]) <= 1 ), name="shift_start_max")
	# OPTION 3: SOS constraint
	#for b in burgermakers: 
	#	m.addSOS(GRB.SOS_TYPE1, [v[(b,t)] for t in timeperiods])

	# Add constraints to make sure Barb and Bri work at the same time
	# and Bri and Jack work at separate times
	m.addConstrs((not_s[(b,t)] == (1 - s[(b,t)]) for b in burgermakers for t in timeperiods), name="opposite_s")
	# OPTION 1: using the AND in a general constraint
	#for t in timeperiods: 
	#	m.addGenConstrAnd(s[('Bri',t)], [s[('Barb',t)], not_s[('Jack',t)]], name="policies")
	# OPTION 2: using indicator constraints
	for t in timeperiods: 
		m.addGenConstrIndicator(s[('Barb',t)], True, s[('Bri',t)], sense=GRB.EQUAL, rhs=1.0, name="policies1")
		m.addGenConstrIndicator(s[('Jack',t)], True, s[('Bri',t)], sense=GRB.EQUAL, rhs=0.0, name="policies2")

	# Add constraints so that either Tom or Bri is working at all times
	for t in timeperiods: 
		m.addGenConstrOr(man_s[t], [s[('Bri',t)], s[('Tom',t)]], name="manager_schedule")
	m.addConstrs((man_s[t] == 1 for t in timeperiods), name="authority")

	# auto-tune the parameters
	m.tune()

	# optimize!
	m.optimize()

	# write the LP file (optional)
	#m.write('burgers.lp')
	
	# print out the objective function value
	print('Obj: %g' % m.objVal)

	# print out the schedules
	for b in burgermakers:
		print(b)
		x = []
		for t in timeperiods:
			x.append(int(s[(b,t)].X))
		print(x)

# : exceptions :
except GurobiError as e:
	print('Error code ' + str(e.errno) + ": " + str(e))

except AttributeError:
	print('Encountered an attribute error')
